/*
 * gimmix-arith.c -- University of Applied Sciences Giessen
 *                   MMIX-ARITH implementation
 */


#include <stdio.h>
#include <stdlib.h>
#include <string.h>
#include <ctype.h>

#include "gimmix-arith.h"


//#define sign_bit	((unsigned) 0x80000000)

#define ROUND_OFF	1
#define ROUND_UP	2
#define ROUND_DOWN	3
#define ROUND_NEAR	4


//<Stuff for C preprocessor 2>
/* ATTENTION: We assume a standard C compiler. */
/* Then there is nothing to do here. */


//typedef enum {
//  false, true
//} bool;


//<Tetrabyte and octabyte type definitions 3>
//typedef unsigned int tetra;	/* a tetrabyte */
//typedef struct {
//  tetra h;
//  tetra l;
//} octa;				/* an octabyte */


//<Other type definitions 36>


/**************************************************************/


//<Global variables 4>
octa zero_octa = { 0, 0 };		/* 0 */
octa neg_one = { -1, -1 };		/* -1 */
octa inf_octa = { 0x7ff00000, 0 };	/* floating point +infinity */
octa standard_NaN = { 0x7ff80000, 0 };	/* floating point NaN(.5) */
octa aux;				/* auxiliary output of a subroutine */
bool overflow;				/* set by certain subroutines */
					/* for signed arithmetic */

int cur_round;				/* the current rounding mode */


/**************************************************************/

/* addition and subtraction */


//<Subroutines 5>
octa oplus(octa y, octa z) {
  octa x;

  x.h = y.h + z.h;
  x.l = y.l + z.l;
  if (x.l < y.l) {
    x.h++;
  }
  return x;
}


octa ominus(octa y, octa z) {
  octa x;

  x.h = y.h - z.h;
  x.l = y.l - z.l;
  if (x.l > y.l) {
    x.h--;
  }
  return x;
}


octa incr(octa y, int delta) {
  octa x;

  x.h = y.h;
  x.l = y.l + delta;
  if (delta >= 0) {
    if (x.l < y.l) {
      x.h++;
    }
  } else {
    if (x.l > y.l) {
      x.h--;
    }
  }
  return x;
}


/**************************************************************/

/* shifting */


octa shift_left(octa y, int s) {
  /* shift left */
  tetra yhl, ylh;

  while (s >= 32) {
    y.h = y.l;
    y.l = 0;
    s -= 32;
  }
  if (s != 0) {
    yhl = y.h << s;
    ylh = y.l >> (32 - s);
    y.h = yhl + ylh;
    y.l <<= s;
  }
  return y;
}


octa shift_right(octa y, int s, int u) {
  /* shift right, arithmetically if u = 0 */
  tetra yhl, ylh;

  while (s >= 32) {
    y.l = y.h;
    y.h = (u ? 0 : -(y.h >> 31));
    s -= 32;
  }
  if (s != 0) {
    yhl = y.h << (32 - s);
    ylh = y.l >> s;
    y.h = (u ? 0 : (-(y.h >> 31)) << (32 - s)) + (y.h >> s);
    y.l = yhl + ylh;
  }
  return y;
}


/**************************************************************/

/* multiplication */


octa omult(octa y, octa z) {
  int i, j, k;
  tetra u[4], v[4], w[8];
  tetra t;
  octa acc;

  //<Unpack the multiplier and multiplicand to u and v 10>
  u[3] = y.h >> 16;
  u[2] = y.h & 0xffff;
  u[1] = y.l >> 16;
  u[0] = y.l & 0xffff;
  v[3] = z.h >> 16;
  v[2] = z.h & 0xffff;
  v[1] = z.l >> 16;
  v[0] = z.l & 0xffff;
  for (j = 0; j < 4; j++) {
    w[j] = 0;
  }
  for (j = 0; j < 4; j++) {
    if (v[j] == 0) {
      w[j + 4] = 0;
    } else {
      for (i = k = 0; i < 4; i++) {
        t = u[i] * v[j] + w[i + j] + k;
        w[i + j] = t & 0xffff;
        k = t >> 16;
      }
      w[j + 4] = k;
    }
  }
  //<Pack w into the outputs aux and acc 11>
  aux.h = (w[7] << 16) + w[6];
  aux.l = (w[5] << 16) + w[4];
  acc.h = (w[3] << 16) + w[2];
  acc.l = (w[1] << 16) + w[0];
  return acc;
}


octa signed_omult(octa y, octa z) {
  octa acc;

  acc = omult(y, z);
  if (y.h & sign_bit) {
    aux = ominus(aux, z);
  }
  if (z.h & sign_bit) {
    aux = ominus(aux, y);
  }
  overflow = (aux.h != aux.l ||
              (aux.h ^ (aux.h >> 1) ^ (acc.h & sign_bit)));
  return acc;
}


/**************************************************************/

/* division */


octa odiv(octa x, octa y, octa z) {
  /* 2^64*x + y = q*z + r, q returned, r in aux */
  int i, j, k, n, d;
  tetra u[8], v[4], q[4], mask, qhat, rhat, vh, vmh;
  tetra t;
  octa acc;

  //<Check that x < z; otherwise give trivial answer 14>
  if (x.h > z.h ||
      (x.h == z.h && x.l >= z.l)) {
    aux = y;
    return x;
  }
  //<Unpack the dividend and divisor to u and v 15>
  u[7] = x.h >> 16;
  u[6] = x.h & 0xffff;
  u[5] = x.l >> 16;
  u[4] = x.l & 0xffff;
  u[3] = y.h >> 16;
  u[2] = y.h & 0xffff;
  u[1] = y.l >> 16;
  u[0] = y.l & 0xffff;
  v[3] = z.h >> 16;
  v[2] = z.h & 0xffff;
  v[1] = z.l >> 16;
  v[0] = z.l & 0xffff;
  //<Determine the number of significant places n in the divisor v 16>
  for (n = 4; v[n - 1] == 0; n--) ;
  //<Normalize the divisor 17>
  // shift u and v left by d places, until 2^15 <= v[n-1] < 2^16
  vh = v[n - 1];
  for (d = 0; vh < 0x8000; d++, vh <<= 1) ;
  for (j = k = 0; j < n + 4; j++) {
    t = (u[j] << d) + k;
    u[j] = t & 0xffff;
    k = t >> 16;
  }
  for (j = k = 0; j < n; j++) {
    t = (v[j] << d) + k;
    v[j] = t & 0xffff;
    k = t >> 16;
  }
  vh = v[n - 1];
  vmh = (n > 1 ? v[n - 2] : 0);
  // now divide
  for (j = 3; j >= 0; j--) {
    //<Determine the quotient digit q[j] 20>
    //<Find the trial quotient, qhat 21>
    t = (u[j + n] << 16) + u[j + n - 1];
    qhat = t / vh;
    rhat = t - vh * qhat;
    while (qhat == 0x10000 ||
           qhat * vmh > (rhat << 16) + u[j + n - 2]) {
      qhat--;
      rhat += vh;
      if (rhat >= 0x10000) {
        break;
      }
    }
    //<Subtract b^j*qhat*v from u 22>
    for (i = k = 0; i < n; i++) {
      t = u[i + j] + 0xffff0000 - k - qhat * v[i];
      u[i + j] = t & 0xffff;
      k = 0xffff - (t >> 16);
    }
    //<If the result was negative, decrease qhat by 1 23>
    // seldom necessary, example: 0x7fff800100000000 / 0x800080020005
    if (u[j + n] != k) {
      qhat--;
      for (i = k = 0; i < n; i++) {
        t = u[i + j] + v[i] + k;
        u[i + j] = t & 0xffff;
        k = t >> 16;
      }
    }
    // store digit
    q[j] = qhat;
  }
  //<Unnormalize the remainder 18>
  mask = (1 << d) - 1;
  for (j = 3; j >= n; j--) {
    u[j] = 0;
  }
  for (k = 0; j >= 0; j--) {
    t = (k << 16) + u[j];
    u[j] = t >> d;
    k = t & mask;
  }
  //<Pack q and u to acc and aux 19>
  acc.h = (q[3] << 16) + q[2];
  acc.l = (q[1] << 16) + q[0];
  aux.h = (u[3] << 16) + u[2];
  aux.l = (u[1] << 16) + u[0];
  return acc;
}


octa signed_odiv(octa y, octa z) {
  octa yy, zz, q;
  int sy, sz;

  if (y.h & sign_bit) {
    sy = 2;
    yy = ominus(zero_octa, y);
  } else {
    sy = 0;
    yy = y;
  }
  if (z.h & sign_bit) {
    sz = 1;
    zz = ominus(zero_octa, z);
  } else {
    sz = 0;
    zz = z;
  }
  q = odiv(zero_octa, yy, zz);
  overflow = false;
  switch (sy + sz) {
    case 2+1:
      aux = ominus(zero_octa, aux);
      if (q.h == sign_bit) {
        overflow = true;
      }
      /* falls through */
    case 0+0:
      return q;
    case 2+0:
      if (aux.h != 0 || aux.l != 0) {
        aux = ominus(zz, aux);
      }
      goto negate_q;
    case 0+1:
      if (aux.h != 0 || aux.l != 0) {
        aux = ominus(aux, zz);
      }
negate_q:
      if (aux.h != 0 || aux.l != 0) {
        return ominus(neg_one, q);
      } else {
        return ominus(zero_octa, q);
      }
    default:
      fprintf(stderr, "Error: This should never happen -- signed_odiv()\n");
      exit(-99);
  }
}


/**************************************************************/

/* bit fiddling */


octa oand(octa y, octa z) {
  octa x;

  x.h = y.h & z.h;
  x.l = y.l & z.l;
  return x;
}


octa oandn(octa y, octa z) {
  octa x;

  x.h = y.h & ~z.h;
  x.l = y.l & ~z.l;
  return x;
}


octa oxor(octa y, octa z) {
  octa x;

  x.h = y.h ^ z.h;
  x.l = y.l ^ z.l;
  return x;
}


int count_bits(tetra x) {
  /* Gillies-Miller method for sideways addition */
  int xx;

  xx = x;
  xx = (xx & 0x55555555) + ((xx >> 1) & 0x55555555);
  xx = (xx & 0x33333333) + ((xx >> 2) & 0x33333333);
  xx = (xx + (xx >> 4)) & 0x0f0f0f0f;
  xx = xx + (xx >> 8);
  return (xx + (xx >> 16)) & 0xff;
}


tetra byte_diff(tetra y, tetra z) {
  tetra d, m, x;

  d = (y & 0x00ff00ff) + 0x01000100 - (z & 0x00ff00ff);
  m = d & 0x01000100;
  x = d & (m - (m >> 8));
  d = ((y >> 8) & 0x00ff00ff) + 0x01000100 - ((z >> 8) & 0x00ff00ff);
  m = d & 0x01000100;
  return x + ((d & (m - (m >> 8))) << 8);
}


tetra wyde_diff(tetra y, tetra z) {
  tetra a, b;

  a = ((y >> 16) - (z >> 16)) & 0x10000;
  b = ((y & 0xffff) - (z & 0xffff)) & 0x10000;
  return y - (z ^ ((y ^ z) & (b - a - (b >> 16))));
}


octa bool_mult(octa y, octa z, bool xor) {
  octa o, x;
  tetra a, b, c;
  int k;

  for (k = 0, o = y, x = zero_octa;
       o.h != 0 || o.l != 0;
       k++, o = shift_right(o, 8, 1)) {
    if (o.l & 0xff) {
      a = ((z.h >> k) & 0x01010101) * 0xff;
      b = ((z.l >> k) & 0x01010101) * 0xff;
      c = (o.l & 0xff) * 0x01010101;
      if (xor) {
        x.h ^= a & c;
        x.l ^= b & c;
      } else {
        x.h |= a & c;
        x.l |= b & c;
      }
    }
  }
  return x;
}


/**************************************************************/

/* floating point packing and unpacking */
