/*
 * gimmotype.c -- University of Applied Sciences Giessen
 *                MMOTYPE implementation
 */


#include <stdio.h>
#include <stdlib.h>
#include <string.h>
#include <time.h>

#include "gimmix-common.h"


#define err(m)	{ fprintf(stderr, "Error in tetra %d: %s!\n", count, m);\
		  continue; }

#define y	buf[2]		/* the next-to-least significant byte */
#define z	buf[3]		/* the least significant byte */

#define sym_length_max	1000	/* max length of symbols in symbol table */


//<Global variables 4>
int listing;		/* are we listing everything? */
int verbose;		/* are we also showing the tetras of input? */
FILE *mmo_file;		/* the input file */
int count;		/* the number of tetrabytes we've read */
int byte_count;		/* index of the next-to-be-read byte */
byte buf[4];		/* the most recently read bytes */
int yz;			/* the two least significant bytes */
tetra tet;		/* buf bytes packed big-endianwise */
octa cur_loc;		/* the current location */
int listed_file;	/* the most recently listed file number */
int cur_file;		/* the most recently selected file number */
int cur_line;		/* the current position in cur_file */
char *file_name[256];	/* file names seen */
octa tmp;		/* an octabyte of temporary interest */
int stab_start;		/* where the symbol table began */
char sym_buf[sym_length_max];	/* the characters on middle
				   transitions to current node */
char *sym_ptr;			/* the character in sym_buf
				   following the current prefix */
char equiv_buf[20];		/* equivalent of the current symbol */


//<Subroutines 8>
octa incr(octa o, int delta) {
  tetra t;
  octa x;

  if (delta >= 0) {
    t = 0xffffffff - delta;
    if (o.l <= t) {
      x.l = o.l + delta;
      x.h = o.h;
    } else {
      x.l = o.l - t - 1;
      x.h = o.h + 1;
    }
  } else {
    t = -delta;
    if (o.l >= t) {
      x.l = o.l - t;
      x.h = o.h;
    } else {
      x.l = o.l + (0xffffffff + delta) + 1;
      x.h = o.h - 1;
    }
  }
  return x;
}


void read_tet(void) {
  if (fread(buf, 1, 4, mmo_file) != 4) {
    fprintf(stderr, "Unexpected end of file after %d tetras!\n", count);
    exit(-3);
  }
  yz = (buf[2] << 8) + buf[3];
  tet = (((buf[0] << 8) + buf[1]) << 16) + yz;
  if (verbose) {
    printf("  %08x\n", tet);
  }
  count++;
}


byte read_byte(void) {
  byte b;

  if (byte_count == 0) {
    read_tet();
  }
  b = buf[byte_count];
  byte_count = (byte_count + 1) & 3;
  return b;
}


void print_stab(void) {
  int m;  /* the master control byte */
  int c;  /* the character at the current trie node */
  int j, k;

  m = read_byte();
  if (m & 0x40) {
    /* traverse the left subtrie */
    print_stab();
  }
  if (m & 0x2f) {
    //<Read the character c 27>
    if (m & 0x80) {
      /* 16-bit character */
      j = read_byte();
    } else {
      j = 0;
    }
    c = read_byte();
    if (j != 0) {
      /* oops, we can't print (j << 8) + c easily at this time */
      c = '?';
    }
    *sym_ptr++ = c;
    if (sym_ptr == &sym_buf[sym_length_max]) {
      fprintf(stderr, "Oops, the symbol is too long!\n");
      exit(-7);
    }
    if (m & 0xf) {
      //<Print the current symbol with its equivalent and serial number 28>
      *sym_ptr = '\0';
      j = m & 0xf;
      if (j == 15) {
        sprintf(equiv_buf, "$%03d", read_byte());
      } else
      if (j <= 8) {
        strcpy(equiv_buf, "#");
        for (; j > 0; j--) {
          sprintf(equiv_buf + strlen(equiv_buf), "%02x", read_byte());
        }
        if (strcmp(equiv_buf, "#0000") == 0) {
          /* undefined */
          strcpy(equiv_buf, "?");
        }
      } else {
        strncpy(equiv_buf, "#20000000000000", 33 - 2 * j);
        equiv_buf[33 - 2 * j] = '\0';
        for (; j > 8; j--) {
          sprintf(equiv_buf + strlen(equiv_buf), "%02x", read_byte());
        }
      }
      for (j = k = read_byte(); ; k = read_byte(), j = (j << 7) + k) {
        if (k >= 128) {
          /* the serial number is now j - 128 */
          break;
        }
      }
      printf("    %s = %s (%d)\n", sym_buf + 1, equiv_buf, j - 128);
    }
    if (m & 0x20) {
      /* traverse the middle subtrie */
      print_stab();
    }
    sym_ptr--;
  }
  if (m & 0x10) {
    /* traverse the right subtrie */
    print_stab();
  }
}


int main(int argc, char *argv[]) {
  int j, delta, postamble;
  char *p;

  postamble = 0;
  //<Process the command line 2>
  listing = 1;
  verbose = 0;
  for (j = 1; j < argc - 1 && argv[j][0] == '-' && argv[j][2] == '\0'; j++) {
    if (argv[j][1] == 's') {
      listing = 0;
    } else
    if (argv[j][1] == 'v') {
      verbose = 1;
    } else {
      break;
    }
  }
  if (j != argc - 1) {
    fprintf(stderr, "Usage: %s [-s] [-v] mmofile\n", argv[0]);
    exit(-1);
  }
  //<Initialize everything 3>
  mmo_file = fopen(argv[argc - 1], "rb");
  if (mmo_file == NULL) {
    fprintf(stderr, "Can't open file %s!\n", argv[argc - 1]);
    exit(-2);
  }
  count = 0;
  byte_count = 0;
  cur_loc.h = 0;
  cur_loc.l = 0;
  listed_file = -1;
  cur_file = -1;
  cur_line = 0;
  //<List the preamble 23>
  read_tet();
  if (buf[0] != mm || buf[1] != lop_pre) {
    fprintf(stderr, "Input is not an MMO file (first two bytes are wrong)!\n");
    exit(-5);
  }
  if (y != 1) {
    fprintf(stderr,
            "Warning: I'm reading this file as version 1, not version %d!", y);
  }
  if (z > 0) {
    j = z;
    read_tet();
    if (listing) {
      printf("File was created %s", asctime(localtime((time_t *) &tet)));
    }
    for (j--; j > 0; j--) {
      read_tet();
      if (listing) {
        printf("Preamble data %08x\n", tet);
      }
    }
  }
  do {
    //<List the next item 13>
    read_tet();
loop:
    if (buf[0] == mm) {
      switch (buf[1]) {
        case lop_quote:
          if (yz != 1) {
            err("YZ field of lop_quote should be 1");
          }
          read_tet();
          break;
        //<Cases for lopcodes in the main loop 18>
        case lop_loc:
          if (z == 2) {
            j = y;
            read_tet();
            cur_loc.h = (j << 24) + tet;
          } else
          if (z == 1) {
            cur_loc.h = y << 24;
          } else {
            err("Z field of lop_loc should be 1 or 2");
          }
          read_tet();
          cur_loc.l = tet;
          continue;
        case lop_skip:
          cur_loc = incr(cur_loc, yz);
          continue;
        case lop_fixo:
          if (z == 2) {
            j = y;
            read_tet();
            tmp.h = (j << 24) + tet;
          } else
          if (z == 1) {
            tmp.h = y << 24;
          } else {
            err("Z field of lop_fixo should be 1 or 2");
          }
          read_tet();
          tmp.l = tet;
          if (listing) {
            printf("%08x%08x: %08x%08x\n", tmp.h, tmp.l, cur_loc.h, cur_loc.l);
          }
          continue;
        case lop_fixr:
          delta = yz;
          tmp = incr(cur_loc,
                     -(delta >= 0x1000000 ? (delta & 0xffffff) - (1 << j) :
                                            delta) << 2);
          if (listing) {
            printf("%08x%08x: %08x\n", tmp.h, tmp.l, delta);
          }
          continue;
        case lop_fixrx:
          j = yz;
          if (j != 16 && j != 24) {
            err("YZ field of lop_fixrx should be 16 or 24");
          }
          read_tet();
          delta = tet;
          if (delta & 0xfe000000) {
            err("increment of lop_fixrx is too large");
          }
          tmp = incr(cur_loc,
                     -(delta >= 0x1000000 ? (delta & 0xffffff) - (1 << j) :
                                            delta) << 2);
          if (listing) {
            printf("%08x%08x: %08x\n", tmp.h, tmp.l, delta);
          }
          continue;
        case lop_file:
          if (file_name[y] != NULL) {
            if (z != 0) {
              err("Two file names with the same number");
            }
            for (j = z; j > 0; j--) {
              read_tet();
            }
            cur_file = y;
          } else {
            if (z == 0) {
              err("No name given for newly selected file");
            }
            file_name[y] = (char *) calloc(4 * z + 1, 1);
            if (file_name[y] == NULL) {
              fprintf(stderr, "No room to store the file name!\n");
              exit(-4);
            }
            cur_file = y;
            for (j = z, p = file_name[y]; j > 0; j--, p += 4) {
              read_tet();
              *(p + 0) = buf[0];
              *(p + 1) = buf[1];
              *(p + 2) = buf[2];
              *(p + 3) = buf[3];
            }
          }
          cur_line = 0;
          continue;
        case lop_line:
          if (cur_file < 0) {
            err("No file was selected for lop_line");
          }
          cur_line = yz;
          continue;
        case lop_spec:
          if (listing) {
            printf("Special data %d at loc %08x%08x",
                   yz, cur_loc.h, cur_loc.l);
            if (cur_line == 0) {
              printf("\n");
            } else {
              if (cur_file == listed_file) {
                printf(" (line %d)\n", cur_line);
              } else {
                printf(" (\"%s\", line %d)\n", file_name[cur_file], cur_line);
                listed_file = cur_file;
              }
            }
          }
          while (1) {
            read_tet();
            if (buf[0] == mm) {
              if (buf[1] != lop_quote || yz != 1) {
                /* end of special data */
                goto loop;
              }
              read_tet();
            }
            if (listing) {
              printf("                   %08x\n", tet);
            }
          }
        case lop_pre:
          err("Can't have another preamble");
        case lop_post:
          postamble = 1;
          if (y != 0) {
            err("Y field of lop_post should be zero");
          }
          if (z < 32) {
            err("Z field of lop_post must be 32 or more");
          }
          continue;
        case lop_stab:
          err("Symbol table must follow postamble");
        case lop_end:
          err("Symbol table can't end before it begins");
        default:
          err("Unknown lopcode");
      }
    }
    if (listing) {
      //<List tet as a normal item 15>
      printf("%08x%08x: %08x", cur_loc.h, cur_loc.l, tet);
      if (cur_line == 0) {
        printf("\n");
      } else {
        if (cur_loc.h & 0xe0000000) {
          printf("\n");
        } else {
          if (cur_file == listed_file) {
            printf(" (line %d)\n", cur_line);
          } else {
            printf(" (\"%s\", line %d)\n", file_name[cur_file], cur_line);
            listed_file = cur_file;
          }
        }
        cur_line++;
      }
      cur_loc = incr(cur_loc, 4);
      cur_loc.l &= -4;
    }
  } while (!postamble);
  //<List the postamble 24>
  for (j = z; j < 256; j++) {
    read_tet();
    tmp.h = tet;
    read_tet();
    if (listing) {
      if (tmp.h != 0 || tet != 0) {
        printf("g%03d: %08x%08x\n", j, tmp.h, tet);
      } else {
        printf("g%03d: 0\n", j);
      }
    }
  }
  //<List the symbol table 25>
  read_tet();
  if (buf[0] != mm || buf[1] != lop_stab) {
    fprintf(stderr, "Symbol table does not follow the postamble!\n");
    exit(-6);
  }
  if (yz != 0) {
    fprintf(stderr, "YZ field of lop_stab should be zero!\n");
  }
  printf("Symbol table (beginning at tetra %d):\n", count);
  stab_start = count;
  sym_ptr = sym_buf;
  print_stab();
  //<Check the lop_end 30>
  while (byte_count != 0) {
    if (read_byte() != 0) {
      fprintf(stderr, "Nonzero byte follows the symbol table!\n");
    }
  }
  read_tet();
  if (buf[0] != mm || buf[1] != lop_end) {
    fprintf(stderr, "The symbol table isn't followed by lop_end!\n");
  } else
  if (count != stab_start + yz + 1) {
    fprintf(stderr, "YZ field at lop_end should have been %d!\n",
            count - yz - 1);
  } else {
    if (verbose) {
      printf("Symbol table ends at tetra %d.\n", count);
    }
    if (fread(buf, 1, 1, mmo_file) != 0) {
      fprintf(stderr, "Extra bytes follow the lop_end!\n");
    }
  }
  return 0;
}
